// cclevin class file
//
// clever 1D integration routine based on ADPINT by WJS in FORTRAN
//                              
// The routine and associated subroutines work by splitting up the 
// intervals which seem to contribute the largest error into     
// smaller intervals for integration, repeatedly.
//
// v1.0 by Martin Kimber 20 July 2000 + tiny change 14 May 2001
//                                      to limit to 100 warnings
// Comments to: martin.kimber@physics.org
//

#include <stdio.h>
#include <iostream.h>
#include <math.h>
#include "cclevin.h"

void c_clevin::set_maxint(int num) // maxint is allowed number of
{                                  // intervals for integration
  if(num<20) num=20; if(num>MAX_INT) num=MAX_INT;
  maxint=num; calculated=0;
}

void c_clevin::set_integrand(double (*ptr_func)(double x))
  // set_integrand should be given the name of a global function
  // of type double taking one double argument; this can act as
  // an interface to member functions of a global object
{  
  fct=ptr_func; calculated=0;
}  

void c_clevin::set_aerr(double absolute_error_reqd)
{
  aerr=absolute_error_reqd; calculated=0;
}

void c_clevin::set_rerr(double relative_error_reqd)
{
  rerr=relative_error_reqd; calculated=0;
}

c_clevin::c_clevin() // constructor function
{
  set_maxint(500); set_aerr(0); set_rerr(0.001); // defaults
  numint=0;
}

double c_clevin::get_errest()
{
  if(!calculated) cerr << "ERROR: cclevin has not done integral yet!\n";
  return errest;
}  

int c_clevin::check()
{
  if(!calculated) 
  {
    cerr << "ERROR: cclevin has not done integral yet!\n";
    return -1;
  }  
  else
  {
    if(resultcode) warn();
    return resultcode;  
  }
}

void c_clevin::warn() // private function
{
  static long warn_counter=0;
  if(++warn_counter<100)
  {
    cerr << "WARNING: value given by cclevin is " << total;
    cerr << "\nbut error estimate returned as " << errest;
    cerr << "\nNumber of intervals used was " << numint << "\n";
  }
}

double c_clevin::do_integral(double lower,double upper)
// this is the actual adaptive integration routine
{
  double dx,target,totl; //GW x
  int i,j,we_should_continue;

// u,v are pointers to arrays containing start,end of each interval 
// fu,fv are pointers to arrays containing f(x) at these points, and
// fw is a pointer to an array containing f(x) at midpoints of these

  double u[MAX_INT],v[MAX_INT],fu[MAX_INT],fv[MAX_INT],fw[MAX_INT];

// err points to array of estimated errors for each interval, 
// result points to array of estimated areas for each interval
 
  double err[MAX_INT],result[MAX_INT];

  numint=5; // current number of intervals, starts at minimum 

  resultcode=0; // will be set to 1 if maxint intervals exceeded 
                 
  calculated=1; // after this call, error estimate and result code 
                // checking calls are allowed
                
// first, choose the first numint intervals and evaluate integral 
// and estimated errors for them; this may suffice 

  dx = (upper-lower)/numint;

  u[0] = lower;  v[0] = lower + dx;   // first interval 
  fu[0] = fct(u[0]); fv[0] = fct(v[0]);

  calc_int_and_err(u,v,fu,fv,fw,err,result,0);
    // this does the calculation - all arguments are pointers to arrays
    // except the last one, which tells which interval to work on 

  for(i=1;i<numint;i++)   // NB last one is [numint-1] 
  {
     u[i] = v[i-1]; fu[i] = fv[i-1];  // no need to recalculate 
     v[i] = u[i] + dx;
     if (i==(numint-1)) v[i] = upper;  // to avoid rounding errors 

     fv[i] = fct(v[i]);

     calc_int_and_err(u,v,fu,fv,fw,err,result,i);
  }

// the next thing to do is to total the results and error estimates 
// then if this exceeds the target given, repeatedly call the slave 
// routine split_and_calc until the error is acceptable (unless the 
// number of intervals needed exceeds maxint) 

  we_should_continue=1; // must execute code at least once 
              
  while(we_should_continue)   // this is the main loop 
  {
    totl=0.; // will be the result of integral 
    errest=0.; // will be our estimated total error 

    for(i=0;i<numint;i++) // less than because array starts [0] 
    {
      totl+=result[i];  errest+=err[i];
    }

    target=fabs(aerr) + fabs(rerr*totl);  // want to be <= this 

    if(errest<=target)
    {
      we_should_continue=0; // target met, so end the loop 
      resultcode=0; // in case exactly maxint intervals sufficed 
    }
    else if(resultcode==1)
    {
      we_should_continue=0;  // maxint intervals done, abort 
    }
    else  // we have to split up offending intervals 
    {
      j=numint;  // have to copy as numint will change 

      for(i=0;i<j;i++)   // look at each interval in turn 
      {
        if((err[i]*2.*j) > target)  // then split this one 
           resultcode= split_and_calc(u,v,fu,fv,fw,err,result,i);

         // note: if at any point resultcode is set to 1, that 
         // and subsequent calls to split_and_calc do nothing 
         // and the routine will abort on the next main loop pass 
      }
    }
  }  
  return totl;
}

// private (slave) functions for do_integral:

void c_clevin::calc_int_and_err(double *u,double *v,double *fu,
  double *fv,double *fw,double *err,double *result,int thi)
{
// this routine calculates integral and error using u[thi], v[thi],
// fu[thi], fv[thi], returning fw[thi] (midpoint value), result[thi]
// and err[thi] which is the estimated error 

// Simpson's rule is used 

  double dx;

  fw[thi] = fct(.5*(u[thi] + v[thi]));  // fill in midpoint 
  dx = v[thi] - u[thi];
  result[thi] = dx * (fu[thi] + (4.* fw[thi]) + fv[thi]) / 6.;
  err[thi] = fabs(dx * (fu[thi] -(2.* fw[thi]) +fv[thi]) /12.);
}

int c_clevin::split_and_calc(double *u,double *v,double *fu,
  double *fv,double *fw,double *err,double *result,int thi)
{
// this routine splits interval 'thi' and calls calc_int_and_err
// to produce results for the two new intervals, unless numint would
// exceed maxint in so doing, in which case it does nothing but
// returns a result code of 1 rather than 0 (success)  

  int that; // opposite number to the old variable this renamed 'thi'
  
  // first take care of error case 
  if(numint>=maxint) return 1;  // NB not MAX_INT (user can set maxint)

  // split interval 'thi' in half (NB second half stored at end of
  // the arrays - there is no point in shuffling all the data around) 

  numint++;

  that=numint-1; // this is the array subscript for new interval 

  v[that] = v[thi];    // the new interval 
  fv[that] = fv[thi];
  u[that] = 0.5*(u[thi] + v[thi]);
  fu[that] = fw[thi];  // use previously calculated midpoint 

  v[thi] = u[that];  // old interval now truncated halfway 
  fv[thi] = fw[thi];

  calc_int_and_err(u,v,fu,fv,fw,err,result,thi); // will do fw 

  calc_int_and_err(u,v,fu,fv,fw,err,result,that);

  return 0;
}
