// cParton class implementation
// Comments to: watt(at)hep.ucl.ac.uk
// Based on original code by Jeppe Andersen

#include "parton.h"

extern "C" {
#ifdef PDF_MRST2001LO
  void mrstlo_(double &x,double &mu, size_t &pdfmode,
	       double &upv, double &dnv, double &usea,
	       double &dsea, double &str, double &chm,
	       double &bot, double &glu);
#elif PDF_CTEQ6L1
  void setctq6_(size_t &Iset);
  double ctq6pdf_(int &Iparton, double &X, double &Q);
#elif PDF_CTEQ5L
  void setctq5_(size_t &Iset);
  double ctq5pdf_(int &Iparton, double &X, double &Q);
#elif PDF_CTEQ5LPAR
  double ctq5pd_(size_t &Iset, int  &Iprtn,
		 double &X, double &Q, size_t &Irt);
#elif PDF_GRV94LO
  void grv94lo_(double &X, double &Q2, double &UV, double &DV,
		double &DEL, double &UDB, double &SB, double &GL);
#elif PDF_MRST2002NLO
  void mrst2002_(double &x,double &q,size_t &pdfmode,
		 double &upv,double &dnv,double &usea,
		 double &dsea,double &str,double &chm,
		 double &bot,double &glu);
#endif
}


// constructor function:
cParton::cParton(void)
{
#ifdef PDF_MRST2001LO
  consts.pdfset=MRST2001LO;
  consts.pdfmode=1;      consts.loop=1;
  consts.mcharm = 1.43;  consts.mbottom = 4.3;
  consts.xmin = 1.e-5;   consts.xmax = 1.;
  consts.musqmin = 1.25; consts.musqmax = 1.e7;
  consts.LQCD = 0.220;   consts.aSMZ2 = 0.130;
  //  cout << "Using MRST2001 LO partons..." << endl;
#elif PDF_CTEQ6L1
  consts.pdfset=CTEQ6L1;
  consts.pdfmode=4;      consts.loop=1;
  consts.mcharm = 1.3;   consts.mbottom = 4.5;
  consts.xmin = 1.e-6;   consts.xmax = 1.;
  consts.musqmin = 1.69; consts.musqmax = 1.e8;
  consts.LQCD = 0.215;   consts.aSMZ2 = 0.130;
  setctq6_(consts.pdfmode);
  //  cout << "Using CTEQ6L1 partons..." << endl;
#elif PDF_CTEQ5L
  consts.pdfset=CTEQ5L;
  consts.pdfmode=3;     consts.loop=1;
  consts.mcharm = 1.3;  consts.mbottom = 4.5;
  consts.xmin = 1.e-5;  consts.xmax = 1.;
  consts.musqmin = 1.0; consts.musqmax = 1.e8;
  consts.LQCD = 0.192;  consts.aSMZ2 = 0.127;
  setctq5_(consts.pdfmode);
  //  cout << "Using CTEQ5L partons..." << endl;
#elif PDF_CTEQ5LPAR
  consts.pdfset=CTEQ5LPAR;
  consts.pdfmode=3;      consts.loop=1;
  consts.mcharm = 1.3;   consts.mbottom = 4.5;
  consts.xmin = 1.e-6;   consts.xmax = 1.;
  consts.musqmin = 1.21; consts.musqmax = 1.e6;
  consts.LQCD = 0.192;   consts.aSMZ2 = 0.127;
  //  cout << "Using CTEQ5L (parameterised) partons..." << endl;
#elif PDF_GRV94LO
  consts.pdfset=GRV94LO;
  consts.pdfmode=1;      consts.loop=1;
  consts.mcharm = 1.5;   consts.mbottom = 4.5;
  consts.xmin = 1.e-5;   consts.xmax = 1.;
  consts.musqmin = 0.4; consts.musqmax = 1.e8;
  consts.LQCD = 0.2;     consts.aSMZ2 = 0.128;
  //  cout << "Using GRV94LO partons..." << endl;
#elif PDF_MRST2002NLO
  consts.pdfset=MRST2002NLO;
  consts.pdfmode=1;      consts.loop=2;
  consts.mcharm = 1.43;  consts.mbottom = 4.3;
  consts.xmin = 1.e-5;   consts.xmax = 1.;
  consts.musqmin = 1.25; consts.musqmax = 1.e7;
  consts.LQCD = 0.353;   consts.aSMZ2 = 0.1197;
  //  cout << "Using MRST2002 NLO partons..." << endl;
#else
  error("cParton::cParton","pdfset",consts.pdfset);
#endif
  
  // table points to the parton content:
  table[0]=&content.u;
  table[1]=&content.d;
  table[2]=&content.s;
  table[3]=&content.c;
  table[4]=&content.b;
  table[5]=&content.ubar;
  table[6]=&content.dbar;
  table[7]=&content.g;

  for (int i=0;i<=7;i++) *table[i]=0.; // initialise
}


// accessor function to get consts:
const sPdfConsts cParton::getConsts(void) const { return consts; }


// accessor function to get pointer to content:
const sPdfContent* cParton::getContent(void) const { return &content; }


// accessor function to get pointer to table[]:
const double* cParton::getTable(void) const { return *table; }


// update parton content:
void cParton::update(double x, double musq)
{
  if (x<consts.xmin || x>consts.xmax) {
    warning("cParton::update","x",x);
    if (x<consts.xmin) x = consts.xmin;
    if (x>consts.xmax) x = consts.xmax;
  }
  if (musq<consts.musqmin || musq>consts.musqmax) {
    //    warning("cParton::update","musq",musq);
    if (musq<consts.musqmin) musq = consts.musqmin;
    if (musq>consts.musqmax) musq = consts.musqmax;
  }

#ifndef PDF_GRV94LO
  double mu = sqrt(musq);
#endif

#if PDF_MRST2001LO
  double upv,dnv,usea,dsea,str,chm,bot,glu;
  mrstlo_(x,mu,consts.pdfmode,upv,dnv,usea,dsea,str,chm,bot,glu);
  content.u = upv+usea; content.d = dnv+dsea;
  content.s = str; content.c = chm; content.b = bot;
  content.ubar = usea; content.dbar = dsea;
  content.g = glu;
#elif PDF_CTEQ6L1
  int icteq[]={1,2,3,4,5,-1,-2,0};
  for (int i=0; i <=7; i++)
    *table[i] =  x*ctq6pdf_(icteq[i], x, mu);
#elif PDF_CTEQ5L
  int icteq[]={1,2,3,4,5,-1,-2,0};
  for (int i=0; i <=7; i++)
    *table[i] =  x*ctq5pdf_(icteq[i], x, mu);
#elif PDF_CTEQ5LPAR
  int icteq[]={1,2,3,4,5,-1,-2,0};
  size_t Irt;
  for (int i=0; i <=7; i++) {
    *table[i] = x*ctq5pd_(consts.pdfmode,icteq[i],x,mu,Irt);
    if (Irt!=0) error("cParton::update","Irt",Irt);
  }
#elif PDF_GRV94LO
  double UV, DV, DEL, UDB, SB, GL;
  grv94lo_(x, musq, UV, DV, DEL, UDB, SB, GL);
  content.u = UV + 0.5*(UDB-DEL); content.d = DV + 0.5*(UDB+DEL);
  content.s = SB; content.c = 0.; content.b = 0.;
  content.ubar = 0.5*(UDB-DEL); content.dbar = 0.5*(UDB+DEL);
  content.g = GL;
#elif PDF_MRST2002NLO
  double upv,dnv,usea,dsea,str,chm,bot,glu;
  mrst2002_(x,mu,consts.pdfmode,upv,dnv,usea,dsea,str,chm,bot,glu);
  content.u = upv+usea; content.d = dnv+dsea;
  content.s = str; content.c = chm; content.b = bot;
  content.ubar = usea; content.dbar = dsea;
  content.g = glu;
#else
  error("cParton::update","pdfset",consts.pdfset);
#endif
  
  return;
}


// QCD running coupling at LO or NLO:
double cParton::alpha_S(double musq) const
{
  double mu = sqrt(musq); // renormalisation scale
  static const double mcharm=consts.mcharm;   // charm quark mass
  static const double mbottom=consts.mbottom; // bottom quark mass
  static const double LQCD=consts.LQCD; // Lambda_QCD for nF = 4
  double lambda; // Lambda_QCD for nF = 3, 4 or 5
  double lsq;    // Lambda_QCD squared in GeV^2
  double b;      // LO beta function coefficient for nF
  int nF;        // number of active flavours
  
  if (consts.loop!=1&&consts.loop!=2) error("alpha_S","loop",consts.loop);

  if(mu>=mbottom) {
    nF=5;
    if (consts.loop==2) lambda = 0.249; // MRST2002 NLO only
    else lambda = LQCD * pow( LQCD/mbottom, 2./23.);
  } else if(mu>=mcharm) {
    nF=4;
    lambda = LQCD;
  } else {
    nF=3;
    if (consts.loop==2) lambda = 0.404; // MRST2002 NLO only
    else lambda = LQCD * pow( mcharm/LQCD, 2./27.);
  }
    
  b=(33.-(2.* nF ))/(12.*M_PI);
  lsq = lambda * lambda;

  if (musq<lsq) error("cParton::alpha_s","musq", musq);

  if (consts.loop==1) return(1./(b*log(musq/lsq)));
  else {
    // NLO beta function coefficient for nF:
    double bprime = (153.-19.*nF)/(2.*M_PI*(33.-2.*nF));
    return (1./(b*log(musq/lsq)))
      * ( 1.-(bprime/b)*log(log(musq/lsq))/log(musq/lsq) );
  }
}
