// cUnintParton class implementation.
// Comments to: watt(at)hep.ucl.ac.uk
// Based on original code by Martin Kimber.

#include "parton.h"

// colour factors:
static const double CA=3.;
static const double CF=4./3.;

static const double oneovertwopi=1./(2.*M_PI);



// global object containing integrands (NB referred to via global functions)
class cUnintIntegrands : public cParton
{
public:
  void set_vars(double sx,double skt,double smu,int sf);
  void set_mu(double smu);
  double gluon_Sudakov_integrand(double t);
  double quark_Sudakov_integrand(double t);
  double gluon_integrand(double z);
  double quark_integrand(double z);
private:
  double x,kt,mu; // these are other variables (held constant
  int flavour;    // during 1D integrations, but changed by the
                  // calling routine cUnintParton::calc)
} unint_integrands; // a global object of this class



// global functions (C-style) used by cclevin 1D integration routine
// as interfaces to the integrand functions living in the
// unint_integrands global object

double unint_interface_gluon_Sudakov_integrand(double t)
{ return unint_integrands.gluon_Sudakov_integrand(t); }

double unint_interface_quark_Sudakov_integrand(double t)
{ return unint_integrands.quark_Sudakov_integrand(t); }

double unint_interface_gluon_integrand(double z)
{ return unint_integrands.gluon_integrand(z); }

double unint_interface_quark_integrand(double z)
{ return unint_integrands.quark_integrand(z); }



////////////////////////////////////////
// cUnintParton class implementation: //
////////////////////////////////////////


// Constructor function
cUnintParton::cUnintParton(bool interSud)
{
  interpolate = interSud;
  if (interpolate) {
    // Allocate Sudakov grid dynamically
    // (add extra +1 to allow for numerical inaccuracies
    //  in bilinear interpolation) 
    qSudGrid = new double*[NSUDGRIDKT+1];
    gSudGrid = new double*[NSUDGRIDKT+1];
    for (size_t i=0; i < NSUDGRIDKT+1; i++) {
      qSudGrid[i] = new double[NSUDGRIDMU+1];
      gSudGrid[i] = new double[NSUDGRIDMU+1];
    }
    
    ifstream inFile; ofstream outFile;
    double ktsq,musq;

    // Check if sudakov.dat exists:
    inFile.open("sudakov.dat");
    if (inFile.fail()) {
      cout << "Writing " << NSUDGRIDKT << " x " << NSUDGRIDMU << " Sudakov grid to sudakov.dat ... " << flush;
      outFile.open("sudakov.dat");
      outFile << "# " << NSUDGRIDKT << " x " << NSUDGRIDMU << " grid in ln(ktsq) x ln(musq)" << endl;
      outFile << "# Format: quarkSudakov gluonSudakov" << endl;
      
      interpolate = false;
      // Loop over ktsq and musq spaced logarithmically from consts.musqmin to consts.musqmax
      for (size_t i = 0; i < NSUDGRIDKT; i++) {
	ktsq = exp(log(consts.musqmin) + double(i)/double(NSUDGRIDKT-1.)*(log(consts.musqmax)-log(consts.musqmin)));
	for (size_t j = 0; j < NSUDGRIDMU; j++) {
	  musq = exp(log(consts.musqmin) + double(j)/double(NSUDGRIDMU-1.)*(log(consts.musqmax)-log(consts.musqmin)));
	  qSudGrid[i][j] = quarkSudakov(ktsq,musq); gSudGrid[i][j] = gluonSudakov(ktsq,musq);
	  outFile << qSudGrid[i][j] << " " << gSudGrid[i][j] << endl;
	}
      }
      interpolate = true;
      outFile.close();
      cout << "Done." << endl;
    }

    else {
      cout << "Reading Sudakov grid from sudakov.dat ... " << flush;
      inFile.ignore(100,'\n'); // advance to end of line
      inFile.ignore(100,'\n'); // advance to end of line
      // Loop over ktsq and musq spaced logarithmically from consts.musqmin to consts.musqmax
      for (size_t i = 0; i < NSUDGRIDKT; i++) {
	for (size_t j = 0; j < NSUDGRIDMU; j++) {
	  inFile >> qSudGrid[i][j] >> gSudGrid[i][j];
	}
      }    
      inFile.close();
      cout << "Done." << endl;
    }
  }
  
  // set up 1D integration routine with desired error target
  integrate1D.set_rerr(0.001); // aerr=0 by default
}


// Destructor function
cUnintParton::~cUnintParton(void)
{
  if (interpolate){
    // Deallocate Sudakov grid
    for (size_t i = 0; i <  NSUDGRIDKT+1; i++) {
      delete[] qSudGrid[i];
      delete[] gSudGrid[i];
    }
    delete[] qSudGrid;
    delete[] gSudGrid;
  }
}


// Quark Sudakov form factor
double cUnintParton::quarkSudakov(double ktsq, double musq)
{
  if (ktsq>musq) return 1.;
  if (interpolate) {

    if (ktsq<consts.musqmin||ktsq>consts.musqmax) {
      warning("cUnintParton::quarkSudakov","ktsq",ktsq);
      if (ktsq<consts.musqmin) ktsq = consts.musqmin;
      else if (ktsq>consts.musqmax) ktsq = consts.musqmax;
    }
    if (musq<consts.musqmin||musq>consts.musqmax) {
      if (musq<consts.musqmin) warning("cUnintParton::quarkSudakov","musq",musq);
      if (musq<consts.musqmin) musq = consts.musqmin;
      else if (musq>consts.musqmax) musq = consts.musqmax;
    }

    static const double dlogktsq = (log(consts.musqmax) - log(consts.musqmin))/double(NSUDGRIDKT-1);
    static const double dlogmusq = (log(consts.musqmax) - log(consts.musqmin))/double(NSUDGRIDMU-1);
    double ifrac = (log(ktsq) - log(consts.musqmin))/dlogktsq;
    size_t i = size_t (ifrac); ifrac = ifrac - i;
    double jfrac = (log(musq) - log(consts.musqmin))/dlogmusq;
    size_t j = size_t(jfrac); jfrac = jfrac - j;

    // Bilinear interpolation:
    return (1.-ifrac)*(1.-jfrac)*qSudGrid[i][j] + ifrac*(1.-jfrac)*qSudGrid[i+1][j] 
      + ifrac*jfrac*qSudGrid[i+1][j+1] + (1.-ifrac)*jfrac*qSudGrid[i][j+1];
  }
  else {
    double Sfun;
    
    unint_integrands.set_mu(sqrt(musq)); // set mu
    integrate1D.set_integrand(unint_interface_quark_Sudakov_integrand);
    
    Sfun=integrate1D.do_integral(ktsq,musq); // perform integration
    //    integrate1D.check(); // warn if error reached
    if(Sfun<0) error("cUnintParton::quarkSudakov","Sfun",Sfun);
    
    return exp(-1.*Sfun*oneovertwopi);
  }
}


// Gluon Sudakov form factor
double cUnintParton::gluonSudakov(double ktsq, double musq)
{
  if (ktsq>musq) return 1.;
  if (interpolate) {

    if (ktsq<consts.musqmin||ktsq>consts.musqmax) {
      warning("cUnintParton::gluonSudakov","ktsq",ktsq);
      if (ktsq<consts.musqmin) ktsq = consts.musqmin;
      else if (ktsq>consts.musqmax) ktsq = consts.musqmax;
    }
    if (musq<consts.musqmin||musq>consts.musqmax) {
      if (musq<consts.musqmin) warning("cUnintParton::gluonSudakov","musq",musq);
      if (musq<consts.musqmin) musq = consts.musqmin;
      else if (musq>consts.musqmax) musq = consts.musqmax;
    }
    
    static const double dlogktsq = (log(consts.musqmax) - log(consts.musqmin))/double(NSUDGRIDKT-1);
    static const double dlogmusq = (log(consts.musqmax) - log(consts.musqmin))/double(NSUDGRIDMU-1);
    double ifrac = (log(ktsq) - log(consts.musqmin))/dlogktsq;
    size_t i = size_t (ifrac); ifrac = ifrac - i;
    double jfrac = (log(musq) - log(consts.musqmin))/dlogmusq;
    size_t j = size_t(jfrac); jfrac = jfrac - j;
        
    // Bilinear interpolation:
    return (1.-ifrac)*(1.-jfrac)*gSudGrid[i][j] + ifrac*(1.-jfrac)*gSudGrid[i+1][j]
      + ifrac*jfrac*gSudGrid[i+1][j+1] + (1.-ifrac)*jfrac*gSudGrid[i][j+1];
  }
  else {
    double Sfun;
    
    unint_integrands.set_mu(sqrt(musq)); // set mu
    integrate1D.set_integrand(unint_interface_gluon_Sudakov_integrand);
    
    Sfun=integrate1D.do_integral(ktsq,musq); // perform integration
    //    integrate1D.check(); // warn if error reached
    if(Sfun<0) error("cUnintParton::gluonSudakov","Sfun",Sfun);
        
    return exp(-1.*Sfun*oneovertwopi);
  }
}


// Calculate the unintegrated PDF
double cUnintParton::calc(size_t flavour,double x,double ktsq,double musq)
{
  double answer=0.0,kt=sqrt(ktsq),mu=sqrt(musq),Wfun,upper=1.;
  unint_integrands.set_vars(x,kt,mu,flavour);
  
//   // handle non-perturbative region with simple extension
//   // which ensures normalisation:
//   if(ktsq<consts.musqmin) {
//     if(flavour==7)
//       answer = gluonSudakov(consts.musqmin,musq);
//     else if (flavour<=6)
//       answer = quarkSudakov(consts.musqmin,musq);
//     else
//       error("cUnintParton::calc","flavour",flavour);
//     update(x,consts.musqmin);
//     return answer * (*table[flavour]) * ktsq/consts.musqmin;
//   }
  // New version (06/04/04): ensure both normalisation AND continuity
  static const double ktsqcutoff = consts.musqmin;
  if(ktsq<ktsqcutoff) {
    double sudakovMu0,intMu0,unintMu0;
    if(flavour==7)
      sudakovMu0 = gluonSudakov(ktsqcutoff,musq);
    else if (flavour<=6)
      sudakovMu0 = quarkSudakov(ktsqcutoff,musq);
    else
      error("cUnintParton::calc","flavour",flavour);
    if (x<consts.xmin||x>consts.xmax) return 0.;
    else update(x,ktsqcutoff);
    intMu0 = *table[flavour];
    unintMu0 = calc(flavour,x,ktsqcutoff,musq);
    answer = unintMu0 + 2.*(1.-ktsq/ktsqcutoff)*(intMu0*sudakovMu0-unintMu0);
    if (answer < 0.) answer = 0.; // if negative, set to zero
    return answer * ktsq/ktsqcutoff;
  }

  // first we do the Sudakov factor part
  if(flavour==7)
    answer = gluonSudakov(ktsq,musq);
  else if (flavour<=6)
    answer = quarkSudakov(ktsq,musq);
  else
    error("cUnintParton::calc","flavour",flavour);

   // second we do the last-step evolution part
  
  if(flavour==7) // gluon
    integrate1D.set_integrand(unint_interface_gluon_integrand);
  else // quark
    integrate1D.set_integrand(unint_interface_quark_integrand);

  // Need below for CTEQ6L1 and GRV94LO to exclude x/z=1 (exactly)
  if (consts.pdfset==CTEQ6L1 || consts.pdfset==GRV94LO)
    x *= (1.+1.e-10);
  
  Wfun=integrate1D.do_integral(x,upper); // perform integration
  //  integrate1D.check(); // warn if error reached
  
  answer*=Wfun;

  return answer;
}


// Calculate the doubly-unintegrated PDF
double cUnintParton::calc(size_t flavour,double x,double z,
			  double ktsq,double musq)
{
  double answer,kt=sqrt(ktsq),mu=sqrt(musq);

  unint_integrands.set_vars(x,kt,mu,flavour); // set other vars

//   // Don't handle non-perturbative region:
//   if(ktsq<consts.musqmin)
//     error("cUnintParton::calc","ktsq",ktsq);

  // handle non-perturbative region with simple extension
  // which ensures normalisation AND continuity (06/04/04):
  static const double ktsqcutoff = consts.musqmin;
  if(ktsq<ktsqcutoff) {
    double sudakovMu0,intMu0,unintMu0;
    if(flavour==7)
      sudakovMu0 = gluonSudakov(ktsqcutoff,musq);
    else if (flavour<=6)
      sudakovMu0 = quarkSudakov(ktsqcutoff,musq);
    else
      error("cUnintParton::calc","flavour",flavour);
    if (x<consts.xmin) return 0.;
    else update(x,ktsqcutoff);
    intMu0 = *table[flavour];
    unintMu0 = calc(flavour,x,z,ktsqcutoff,musq);
    answer = unintMu0 + 2.*(1.-ktsq/ktsqcutoff)*(intMu0*sudakovMu0/(1.-x)-unintMu0);
    if (answer < 0.) answer = 0.; // if negative, set to zero
    return answer * ktsq/ktsqcutoff;
  }

  // first we do the Sudakov factor part
  if(flavour==7)
    answer = gluonSudakov(ktsq,musq);
  else if (flavour<=6)
    answer = quarkSudakov(ktsq,musq);
  else
    error("cUnintParton::calc","flavour",flavour);

  // return distribution unintegrated over z:
  if(flavour==7) // gluon
    answer*=unint_interface_gluon_integrand(z);
  else // quark
    answer*=unint_interface_quark_integrand(z);

  return answer;
}



////////////////////////////////////////////
// cUnintIntegrands class implementation: //
////////////////////////////////////////////


inline void cUnintIntegrands::set_vars(double sx,double skt,double smu,int sf)
{ x=sx, kt=skt, mu=smu, flavour=sf; }


inline void cUnintIntegrands::set_mu(double smu)
{ mu=smu; }


// gluon_Sudakov_integrand: gluon Sudakov factor (before exponentiation)
double cUnintIntegrands::gluon_Sudakov_integrand(double t) // t = kappat^2
{
  double k,zmax,zsq,zcub,s1,rest;
  int nF; // number of qqbar flavours gluon can split to
  
  k=sqrt(t);
  if(k>=consts.mbottom) nF=5;
  else if(k>=consts.mcharm) nF=4; 
  else nF=3;

  rest=alpha_S(t)/t;
  zmax=mu/(mu+k); zsq=zmax*zmax; zcub=zsq*zmax;
  //[GW,02/10/02] No angular-ordered quark emission:
  s1=nF/3.; // P_qg part
  s1+=CA*(-2./3.*zcub+zsq-2.*log(1.-zmax)+11./6.-4.*zmax+2.*log(zmax));

  return (s1*rest);
}


// quark_Sudakov_integrand: quark Sudakov factor (before exponentiation)
double cUnintIntegrands::quark_Sudakov_integrand(double t) // t = kappat^2
{
  double k,zmax,s2,rest;
  
  rest=alpha_S(t)/t;
  k=sqrt(t); zmax=mu/(mu+k);
  s2=CF*(-zmax -0.5*zmax*zmax -2.*log(1.-zmax));
  return (s2*rest);
}


// unintegrated gluon integrand:
double cUnintIntegrands::gluon_integrand(double z)
{
  double xoz,omz,Wint,Pgq,Pgg;
  
  xoz=x/z;  // single-scale input functions (integrated) at x/z, kt^2
  update(xoz,kt*kt);
  
  omz=1-z;
  Pgq=CF*(1+(omz*omz))/z;
  Pgg=2*CA*((z*omz)+(omz/z)+(z/omz));
  
  // all flavours of quark, plus all flavours of antiquark, can feed
  // separately into the gluon density evolution 
  
  Wint=0.;
  
  // u=0, d=1, s=2, c=3, b=4, ubar=5, dbar=6:
  for(int partonflav=0;partonflav<=6;partonflav++) 
    Wint+=(*table[partonflav]);
  // sbar=2, cbar=3, bbar=4:
  for(int partonflav=2;partonflav<=4;partonflav++)
    Wint+=(*table[partonflav]);
  
  Wint*=Pgq;
  
  if (z<(mu/(mu+kt)))
    Wint+=Pgg*content.g; // gluon -> gluon splitting
  
  return Wint*alpha_S(kt*kt)*oneovertwopi;
}


// unintegrated quark integrand:
double cUnintIntegrands::quark_integrand(double z)
{
  double xoz,omz,Wint=0.0,Pqg,Pqq;
  omz=1.-z;
  xoz=x/z;   // single-scale input functions (integrated) at x/z, kt
  update(xoz,kt*kt);
  
  omz=1.-z;
  Pqg=0.5*((z*z)+(omz*omz));
  Pqq=CF*(1.+(z*z))/omz;
  
  if (z<(mu/(mu+kt))) //[GW] angular ordered upper limit on z integration
    Wint=Pqq*(*table[flavour]);

  // gluon -> quark splittings only for u,d,s (and antiq) 
  // unless kt is greater than mcharm in which case c allowed
  // this is consistent with new nF treatment in gluon_Sudakov_integrand
  // a similar step happens for bottom

  if((flavour!=3 && flavour!=4) 
     || (flavour==3 && kt>=consts.mcharm)
     || (flavour==4 && kt>=consts.mbottom))
    Wint+=Pqg*content.g;

  return Wint*alpha_S(kt*kt)*oneovertwopi;
}
